// Copyright 2019-2024 Yuji Tanaka
// HP : http://galaxystar.image.coocan.jp/
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
/*
 * 3D shoot
 * 
 * 2019' 08/17 Yuji Tanaka
 * 2024' 05/22 Yuji Tanaka
 *   For ESP32-S3.
 */
#include <SPI.h>
#include <ILI9163_buf.h>
#include <driver/i2s.h>
#include "pitches.h"

#define SD_CS  10
#define TFT_DC 48
#define TFT_CS 47
#define TFT_RST 45

ILI9163_buf tft = ILI9163_buf(TFT_CS, TFT_DC, TFT_RST);

int _width, _height;

int StickXPin = 6;
int StickYPin = 7;
int StickX;
int StickY;

int ButtonAPin = 15;
int ButtonBPin = 16;
int ButtonA;
int ButtonB;
int OldA, OldB;

// Sound
#define fs 11025
#define BLOCK_SAMPLES 256

int SoundPin=0; // Dummy
int rxbuf[BLOCK_SAMPLES*2], txbuf[BLOCK_SAMPLES*2];
float Lch_out[BLOCK_SAMPLES*8], Rch_out[BLOCK_SAMPLES*8];

#define SCREEN_WIDTH  128
#define SCREEN_HEIGHT 128
#define SCREEN_START_X 0
#define SCREEN_START_Y 32
#define MAX_TEKI 16
#define MAX_SHOT 8
#define MAX_BOMB 8

int teki_x[MAX_TEKI], teki_y[MAX_TEKI], teki_z[MAX_TEKI], teki_typ[MAX_TEKI], teki_dir[MAX_TEKI];
int shot_x[MAX_SHOT], shot_y[MAX_SHOT], shot_z[MAX_SHOT];
int bomb_x[MAX_BOMB], bomb_y[MAX_BOMB], bomb_z[MAX_BOMB];

struct POINT {
  int x, y;
};

struct f_vec {
  double x,y,z;
};

struct f_wall{
  double x,y,z;
  int typ;  // 0:point  1:color  -1:end
};

struct f_wall *field1;
struct f_wall *teki1;
struct f_wall *teki2;
struct f_wall *teki3;
struct f_wall *teki4;
struct f_wall *teki5;
struct f_wall *building1;
struct f_wall *building2;
struct f_wall *building3;
struct f_wall *shot;
struct f_wall *bomb;

struct f_wall *wallpt[16];

/*  polygonデータ  */
struct f_polygon {
  POINT pnt[8];
  int n;
  uint16_t c;
  double dmax;
  double dmin;
}*polygon_dat;
int poly_num;

void push_polygon(POINT *, int, uint16_t, double, double);
void draw_polygon();
struct f_vec *pos3d(struct f_vec *,struct f_vec *,struct f_vec *);
struct f_vec *vec_add(struct f_vec *,struct f_vec *,struct f_vec *);
struct f_vec *vec_sub(struct f_vec *,struct f_vec *,struct f_vec *);

int stage1[]= {
// time,   x,   y,   z, type
     10,   0, 800,   0, 8,
     50,  64, 800, 128, 1,
    100, 226, 800,   0, 9,
    150, 192, 800, 192, 1,
    200,  64, 800,   0, 8,
    240,  96, 800, 128, 2,
    260, 160, 800, 128, 2,
    300, 192, 800,   0, 9,
    350, 128, 800, 128, 3,
    400,  30, 800,   0, 8,
    400, 160, 800,   0, 8,
    450,  16, 800, 224, 1,
    450, 240, 800, 192, 1,
    500, 192, 800, 160, 2,
    500,  64, 800, 128, 2,
    550,  96, 800,  96, 4,
    550, 160, 800,  64, 4,
    600, 128, 800, 128, 4,
  65536,   0,   0,   0, 0
};

int stage2[]= {
// time,   x,   y,   z, type
      1, 128, 800,   0, 9,
     50,  64, 800, 128, 2,
     50, 128, 800, 128, 2,
     50, 192, 800, 128, 2,
    100,  20, 800,   0, 8,
    100,  96, 800, 236, 1,
    120, 160, 800, 216, 1,
    140, 160, 800, 196, 1,
    150, 176, 800,   0, 8,
    160, 160, 800, 176, 1,
    200, 160, 800,   0, 9,
    250,  96, 800,   0, 9,
    260, 128, 800, 128, 2,
    280, 128, 800, 192, 2,
    300, 128, 800,   0, 9,
    360,  32, 800, 192, 4,
    380, 224, 800, 192, 4,
    400,  88, 800,   0, 8,
    400,  20, 800, 236, 3,
    400, 236, 800, 236, 3,
    450, 128, 800, 196, 4,
    450, 128, 800,  64, 4,
    500, 128, 800,   0, 9,
    500, 192, 800, 236, 1,
    500,  64, 800, 216, 1,
    500,  32, 800, 196, 1,
    500, 224, 800, 176, 1,
    550,  64, 800, 128, 4,
    550, 192, 800, 128, 4,
  65536,   0,   0,   0, 0
};

int stage3[]= {
// time,   x,   y,   z, type
      1, 192, 800,   0, 8,
     50,  64, 800,  64, 1,
     50, 128, 800, 128, 1,
     50, 192, 800, 192, 1,
    100,  30, 800,   0, 9,
    100,  96, 800,  96, 1,
    100, 160, 800, 160, 1,
    150, 128, 800, 192, 3,
    200,  30, 800,   0, 9,
    200, 226, 800,   0, 9,
    220, 128, 800,  96, 2,
    240, 128, 800, 160, 2,
    260, 128, 800, 128, 2,
    280, 128, 800, 192, 2,
    300, 128, 800,   0, 9,
    320,  64, 800, 128, 3,
    320, 192, 800, 128, 3,
    340,  96, 800, 192, 3,
    340, 160, 800, 192, 3,
    360,  80, 800,  96, 3,
    360, 176, 800,  96, 3,
    380,  20, 800, 236, 3,
    380, 236, 800, 236, 3,
    400,  88, 800,   0, 8,
    450,  16, 800, 236, 1,
    450, 240, 800, 216, 1,
    450, 128, 800, 196, 1,
    500,  64, 800,   0, 10,
    500, 192, 800,   0, 10,
    500, 192, 800, 160, 2,
    500,  64, 800, 128, 2,
    550,  32, 800, 128, 4,
    550, 224, 800, 128, 4,
    575,  64, 800, 128, 4,
    575, 192, 800, 128, 4,
    600, 128, 800,   0, 10,
    620,  96, 800, 192, 4,
    620, 160, 800, 192, 4,
    640,  20, 800, 236, 3,
    640, 236, 800, 236, 3,
  65536,   0,   0,   0, 0
};

int stage4[]= {
// time,   x,   y,   z, type
      1, 236, 800,   0, 10,
      1,  20, 800,   0, 10,
     50,  64, 800, 128, 5,
     50, 128, 800, 128, 5,
     50, 192, 800, 128, 5,
    100,  88, 800,   0, 8,
    100,  96, 800,  96, 1,
    100, 160, 800, 160, 1,
    150, 128, 800, 192, 4,
    150, 192, 800, 128, 4,
    200, 128, 800,   0, 10,
    220,  20, 800, 236, 3,
    240, 236, 800,  20, 3,
    260,  20, 800,  20, 3,
    280, 236, 800, 236, 3,
    300, 128, 800,   0, 9,
    320,  64, 800, 128, 5,
    340, 192, 800, 128, 5,
    360,  96, 800, 192, 5,
    380, 160, 800, 192, 5,
    400,   0, 800,   0, 8,
    450,  16, 800, 236, 1,
    450, 240, 800, 216, 1,
    450, 128, 800, 196, 1,
    500, 186, 800,   0, 8,
    500, 192, 800, 160, 4,
    500,  64, 800, 128, 4,
    550,  32, 800, 128, 5,
    550, 224, 800, 128, 5,
    575,  64, 800, 128, 5,
    575, 192, 800, 128, 5,
    600,  64, 800,   0, 10,
    600, 192, 800,   0, 10,
    650,  96, 800, 192, 5,
    650, 160, 800, 192, 5,
    680,  20, 800, 236, 3,
    680, 236, 800, 236, 3,
    700, 128, 800, 128, 3,
  65536,   0,   0,   0, 0
};

int state;
int stage;
int score;
int my_x, my_y;
int start_x,start_y;
int goal_x,goal_y;
int count;
int stage_count;
int *stage_data;

void setup()
{
    int x, y;

    Serial.begin(115200);

    tft.begin(40000000);
    tft.setRotation(2);

// Line inversion
    tft.startWrite();
    tft.writeCommand(0xb4);
    tft.spiWrite(0x00); // Line inversion
    tft.endWrite();

    tft.clearDisplay();
    _width=tft.width();
    _height=tft.height();
    tft.display();

    pinMode(ButtonAPin, INPUT_PULLUP);
    pinMode(ButtonBPin, INPUT_PULLUP);
  
    i2s_init();
    
    state=0;
    stage=1;
    count=0;

    polygon_dat=(f_polygon *)malloc(sizeof(struct f_polygon)*512);

    make_field(-1);
    make_teki1(-1);
    make_teki2(-1);
    make_teki3(-1);
    make_teki4(-1);
    make_teki5(-1);
    make_building1();
    make_building2();
    make_building3();
    make_bomb();
    make_shot();

    wallpt[0]=field1;
    wallpt[1]=teki1;
    wallpt[2]=teki2;
    wallpt[3]=teki3;
    wallpt[4]=teki4;
    wallpt[5]=teki5;
    wallpt[8]=building1;
    wallpt[9]=building2;
    wallpt[10]=building3;
    wallpt[14]=bomb;
    wallpt[15]=shot;
}

void loop(){
    int b, n;

    StickX = analogRead(StickXPin)/4;
    StickY = analogRead(StickYPin)/4;
    ButtonA = digitalRead(ButtonAPin);
    ButtonB = digitalRead(ButtonBPin);

    // Title
    if (state==0){
      make_teki1(count);
      make_teki2(count);
      make_teki3(count);
      make_teki4(count);
      make_teki5(count);
      show_title(count*2);
      score=0;
      my_x=128;
      my_y=128;
      if (ButtonA==0 && OldA==1)
        state=1;

    // Initialize
    }else if (state==1){
      count=0;
      state=2;
      stage_count=0;
      if (stage==1)
        stage_data=stage1;
      else if (stage==2)
        stage_data=stage2;
      else if (stage==3)
        stage_data=stage3;
      else
        stage_data=stage4;
      set_teki(-1, 0, 0, 0);
      set_bomb(-1, 0, 0);
      set_shot(-1, 0, 0);
      make_field(1);
      i2s_tone(SoundPin, NOTE_C5, 100);
      delay(50);
      i2s_tone(SoundPin, NOTE_E5, 100);
      delay(50);
      i2s_tone(SoundPin, NOTE_G5, 100);
      delay(50);

    // Game
    }else if (state==2){
      // Shoot
      if (ButtonA==0 && OldA==1){
        set_shot(my_x, 8, my_y);
        i2s_tone(SoundPin, NOTE_C7, 50);
      }

      // Move
      if (StickX<128)
        my_x-=8;
      else if (StickX<256)
        my_x-=4;
      else if (StickX>896)
        my_x+=8;
      else if (StickX>768)
        my_x+=4;

      if (StickY<128)
        my_y-=8;
      else if (StickY<256)
        my_y-=4;
      else if (StickY>896)
        my_y+=8;
      else if (StickY>768)
        my_y+=4;

      if (my_x<8)
        my_x=8;
      if (my_x>248)
        my_x=248;
      if (my_y<8)
        my_y=8;
      if (my_y>248)
        my_y=248;

      make_teki1(count);
      make_teki2(count);
      make_teki3(count);
      make_teki4(count);
      make_teki5(count);
      n=move_teki();
      b=move_bomb();
      move_shot();

      // Crash
      if (b<0 || n<0){
        state=5;
        i2s_tone(SoundPin, NOTE_C5, 100);
        delay(50);
        i2s_tone(SoundPin, NOTE_G4, 100);
        delay(50);
        i2s_tone(SoundPin, NOTE_E4, 100);
        delay(50);
        i2s_tone(SoundPin, NOTE_C4, 100);
        delay(50);
      }

      // Teki
      while(count==stage_data[stage_count*5]){
        set_teki(stage_data[stage_count*5+1], stage_data[stage_count*5+2], stage_data[stage_count*5+3],
                 stage_data[stage_count*5+4]);
        stage_count++;
      }

      // Clear stage
      if (stage_data[stage_count*5+4]==0 && n==0){
        state=3;
        i2s_tone(SoundPin, NOTE_C6, 100);
        delay(50);
        i2s_tone(SoundPin, NOTE_E6, 100);
        delay(50);
        i2s_tone(SoundPin, NOTE_G6, 100);
        delay(50);
        i2s_tone(SoundPin, NOTE_C7, 100);
        delay(50);
      }

      // Bomb
      if (random(0,5)==0){
        n=random(0,MAX_TEKI);
        if (teki_x[n]>=0 && 1<=teki_typ[n] && teki_typ[n]<=7){
          set_bomb(teki_x[n], teki_y[n], teki_z[n]);
        }
      }
      
      show_game(count*2);
    // Clear
    }else if (state==3){
      show_clear();
      stage++;
      state=4;
    // Next stage
    }else if (state==4){
      if (ButtonB==0 && OldB==1)
        state=1;
    // Game over
    }else if (state==5){
      show_gameover();
      if (ButtonB==0 && OldB==1)
        state=0;
    }

    count++;
    OldA=ButtonA;
    OldB=ButtonB;
    i2s_noTone(SoundPin);
    delay(30);
}

void set_teki(int x, int y, int z, int t)
{
  int i;
  for(i=0;i<MAX_TEKI;i++){
    if (x<0){
      teki_x[i]=-1;
    }else if (teki_x[i]<0){
      teki_x[i]=x;
      teki_y[i]=y;
      teki_z[i]=z;
      teki_typ[i]=t;
      teki_dir[i]=0;
      break;
    }
  }
}

int move_teki()
{
  int i, j;
  int n;

  n=0;
  for(i=0;i<MAX_TEKI;i++){
    if (teki_x[i]>=0){
      // Teki1
      if (teki_typ[i]==1){
        if (teki_y[i]>170)
          teki_y[i]-=4;
        if (teki_dir[i]==0){
          teki_x[i]+=4;
          if (teki_x[i]>226)
            teki_dir[i]=1;
        }else{
          teki_x[i]-=4;
          if (teki_x[i]<30)
            teki_dir[i]=0;
        }
        for(j=0;j<MAX_SHOT;j++){
          if (abs(teki_x[i]-shot_x[j])<=20 && abs(teki_y[i]-shot_y[j])<=20 && abs(teki_z[i]-shot_z[j])<=20){
            score+=30;
            teki_x[i]=-1;
            shot_x[j]=-1;
            i2s_tone(SoundPin, NOTE_C5, 50);
            break;
          }
        }
      // Teki2
      }else if (teki_typ[i]==2){
        if (teki_y[i]>160)
          teki_y[i]-=4;
        if ((teki_dir[i] & 1)==0){
          teki_z[i]+=2;
          if (teki_z[i]>226)
            teki_dir[i]=teki_dir[i] | 1;
        }else{
          teki_z[i]-=2;
          if (teki_z[i]<64)
            teki_dir[i]=teki_dir[i] & 2;
        }
        if ((teki_dir[i] & 2)==0){
          teki_x[i]+=2;
          if (teki_x[i]>226)
            teki_dir[i]=teki_dir[i] | 2;
        }else{
          teki_x[i]-=2;
          if (teki_x[i]<30)
            teki_dir[i]=teki_dir[i] & 1;
        }
        for(j=0;j<MAX_SHOT;j++){
          if (abs(teki_x[i]-shot_x[j])+abs(teki_z[i]-shot_z[j])<=30 && abs(teki_y[i]-shot_y[j])<=20){
            score+=50;
            teki_x[i]=-1;
            shot_x[j]=-1;
            i2s_tone(SoundPin, NOTE_C5, 50);
            break;
          }
        }
      // Teki3
      }else if (teki_typ[i]==3){
        teki_y[i]-=4;
        for(j=0;j<MAX_SHOT;j++){
          if (abs(teki_x[i]-shot_x[j])<=10 && abs(teki_y[i]-shot_y[j])<=20 && abs(teki_z[i]-shot_z[j])<=10){
            score+=300;
            teki_x[i]=-1;
            shot_x[j]=-1;
            i2s_tone(SoundPin, NOTE_C5, 50);
            break;
          }
        }
      // Teki4
      }else if (teki_typ[i]==4){
        if (teki_y[i]>150)
          teki_y[i]-=6;
        if ((teki_dir[i] & 1)==0){
          teki_z[i]+=3;
          if (teki_z[i]>226)
            teki_dir[i]=teki_dir[i] | 1;
        }else{
          teki_z[i]-=3;
          if (teki_z[i]<64)
            teki_dir[i]=teki_dir[i] & 2;
        }
        if ((teki_dir[i] & 2)==0){
          teki_x[i]+=3;
          if (teki_x[i]>226)
            teki_dir[i]=teki_dir[i] | 2;
        }else{
          teki_x[i]-=3;
          if (teki_x[i]<30)
            teki_dir[i]=teki_dir[i] & 1;
        }
        for(j=0;j<MAX_SHOT;j++){
          if (abs(teki_x[i]-shot_x[j])+abs(teki_z[i]-shot_z[j])<=30 && abs(teki_y[i]-shot_y[j])<=20){
            score+=100;
            teki_x[i]=-1;
            shot_x[j]=-1;
            i2s_tone(SoundPin, NOTE_C5, 50);
            break;
          }
        }
      // Teki5
      }else if (teki_typ[i]==5){
        if (teki_y[i]>100)
          teki_y[i]-=6;
        if ((teki_dir[i] & 1)==0){
          teki_z[i]+=4;
          if (teki_z[i]>241)
            teki_dir[i]=teki_dir[i] | 1;
        }else{
          teki_z[i]-=4;
          if (teki_z[i]<15)
            teki_dir[i]=teki_dir[i] & 2;
        }
        if ((teki_dir[i] & 2)==0){
          teki_x[i]+=4;
          if (teki_x[i]>241)
            teki_dir[i]=teki_dir[i] | 2;
        }else{
          teki_x[i]-=4;
          if (teki_x[i]<15)
            teki_dir[i]=teki_dir[i] & 1;
        }
        for(j=0;j<MAX_SHOT;j++){
          if (abs(teki_x[i]-shot_x[j])+abs(teki_z[i]-shot_z[j])<=15 && abs(teki_y[i]-shot_y[j])<=20){
            score+=200;
            teki_x[i]=-1;
            shot_x[j]=-1;
            i2s_tone(SoundPin, NOTE_C5, 50);
            break;
          }
        }
      }else{
        teki_y[i]-=2;
        for(j=0;j<MAX_SHOT;j++){
          if (abs(teki_y[i]-shot_y[j])<=10){
            if (teki_typ[i]==8 && teki_x[i]<=shot_x[j] && shot_x[j]<=teki_x[i]+56 &&
                teki_z[i]<=shot_z[j] && shot_z[j]<=teki_z[i]+56){
              shot_x[j]=-1;
              break;
            }
            if (teki_typ[i]==9 && teki_x[i]-20<=shot_x[j] && shot_x[j]<=teki_x[i]+20 &&
                teki_z[i]<=shot_z[j] && shot_z[j]<=teki_z[i]+80){
              shot_x[j]=-1;
              break;
            }
            if (teki_typ[i]==10 && teki_x[i]-20<=shot_x[j] && shot_x[j]<=teki_x[i]+20 &&
                teki_z[i]<=shot_z[j] && shot_z[j]<=teki_z[i]+256){
              shot_x[j]=-1;
              break;
            }
          }
        }
      }

      // Remove
      if (teki_y[i]<30){
        // Crash
        if (teki_typ[i]==8){
          if (teki_x[i]<=my_x && my_x<=teki_x[i]+56 &&
              teki_z[i]<=my_y && my_y<=teki_z[i]+56)
            n=-100;
        }else if (teki_typ[i]==9){
          if (teki_x[i]-20<=my_x && my_x<=teki_x[i]+20 &&
              teki_z[i]<=my_y && my_y<=teki_z[i]+80)
            n=-100;
        }else if (teki_typ[i]==10){
          if (teki_x[i]-20<=my_x && my_x<=teki_x[i]+20 &&
              teki_z[i]<=my_y && my_y<=teki_z[i]+256)
            n=-100;
        }else if (teki_typ[i]==3){
          if (abs(teki_x[i]-my_x)<=10 && abs(teki_z[i]-my_y)<=10)
            n=-100;
        }else if (abs(teki_x[i]-my_x)<=20 && abs(teki_z[i]-my_y)<=20)
          n=-100;
        if (teki_y[i]<20)
          teki_x[i]=-1;
      }
      n++;
    }
  }
  return(n);
}

void set_bomb(int x, int y, int z)
{
  int i;
  for(i=0;i<MAX_BOMB;i++){
    if (x<0){
      bomb_x[i]=-1;
    }else if (bomb_x[i]<0){
      bomb_x[i]=x;
      bomb_y[i]=y;
      bomb_z[i]=z;
      break;
    }
  }
}

int move_bomb()
{
  int i;
  for(i=0;i<MAX_BOMB;i++){
    if (bomb_x[i]>=0){
      bomb_y[i]-=8;
      if (bomb_y[i]<20){
        // Crash
        if (abs(bomb_x[i]-my_x)<=20 && abs(bomb_z[i]-my_y)<=20)
          return(-1);
        bomb_x[i]=-1;
      }
    }
  }
  return(0);
}

void set_shot(int x, int y, int z)
{
  int i;
  for(i=0;i<MAX_SHOT;i++){
    if (x<0){
      shot_x[i]=-1;
    }else if (shot_x[i]<0){
      shot_x[i]=x;
      shot_y[i]=y;
      shot_z[i]=z;
      break;
    }
  }
}

void move_shot()
{
  int i;
  for(i=0;i<MAX_SHOT;i++){
    if (shot_x[i]>=0){
      shot_y[i]+=8;
      if (shot_y[i]>400)
        shot_x[i]=-1;
    }
  }
}

void set_wall(struct f_wall *wpt, double x, double y, double z, int typ)
{
  wpt->x=x;
  wpt->y=y;
  wpt->z=z;
  wpt->typ=typ;
}

void make_field(int n)
{
    struct f_wall *wpt;

    if (n<0){
      field1=(f_wall *)malloc(sizeof(struct f_wall)*21);
    }
    wpt=field1;

    if (n<=0){
      set_wall(wpt++,  0,  0,  0, 0);
      set_wall(wpt++,256,  0,  0, 0);
      set_wall(wpt++,256,800,  0, 0);
      set_wall(wpt++,  0,800,  0, 0);
      set_wall(wpt++, 64, 32,  0, 1);
      set_wall(wpt++,  0,  0,  0,-1);
    }else if (n==1){
      set_wall(wpt++,  0,  0,  0, 0);
      set_wall(wpt++,256,  0,  0, 0);
      set_wall(wpt++,256,800,  0, 0);
      set_wall(wpt++,  0,800,  0, 0);
      set_wall(wpt++, 64, 32,  0, 1);
      set_wall(wpt++,  0,  0,  0, 0);
      set_wall(wpt++,  0,  0,256, 0);
      set_wall(wpt++,  0,800,256, 0);
      set_wall(wpt++,  0,800,  0, 0);
      set_wall(wpt++, 96, 48,  0, 1);
      set_wall(wpt++,256,  0,  0, 0);
      set_wall(wpt++,256,  0,256, 0);
      set_wall(wpt++,256,800,256, 0);
      set_wall(wpt++,256,800,  0, 0);
      set_wall(wpt++, 32, 16,  0, 1);
      set_wall(wpt++,  0,  0,256, 0);
      set_wall(wpt++,256,  0,256, 0);
      set_wall(wpt++,256,800,256, 0);
      set_wall(wpt++,  0,800,256, 0);
      set_wall(wpt++, 64, 32,  0, 1);
      set_wall(wpt++,  0,  0,  0,-1);
    }
}

void make_teki1(int n)
{
    double ang;
    double fl0, fl1;
    double sz;
    int nn;
    struct f_wall *wpt;

    if (n<0){
      teki1=(f_wall *)malloc(sizeof(struct f_wall)*22);
      return;
    }
    wpt=teki1;

    nn=n % 16;
    ang=M_PI/32.0*(double)nn;
    sz=40;
    fl0=sz*cos(ang);
    fl1=sz*sin(ang);

    set_wall(wpt++, fl1,-fl0,-sz*0.4, 0);
    set_wall(wpt++, fl0, fl1,-sz*0.4, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++,255-nn*8,   0,   0, 1);
    set_wall(wpt++, fl0, fl1,-sz*0.4, 0);
    set_wall(wpt++,-fl1, fl0,-sz*0.4, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++,128-nn*4,   0,   0, 1);
    set_wall(wpt++,-fl1, fl0,-sz*0.4, 0);
    set_wall(wpt++,-fl0,-fl1,-sz*0.4, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++,64+nn*4,   0,   0, 1);
    set_wall(wpt++,-fl0,-fl1,-sz*0.4, 0);
    set_wall(wpt++, fl1,-fl0,-sz*0.4, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++,128+nn*8,   0,   0, 1);
    set_wall(wpt++, fl1,-fl0,-sz*0.4, 0);
    set_wall(wpt++, fl0, fl1,-sz*0.4, 0);
    set_wall(wpt++,-fl1, fl0,-sz*0.4, 0);
    set_wall(wpt++,-fl0,-fl1,-sz*0.4, 0);
    set_wall(wpt++, 64,   0,   0, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void make_teki2(int n)
{
    double ang;
    double fl0x, fl0y, fl1x, fl1y, fl2x, fl2y;
    int c0, c1, c2;
    double sz;
    int nn;
    struct f_wall *wpt;

    if (n<0){
      teki2=(f_wall *)malloc(sizeof(struct f_wall)*25);
      return;
    }
    wpt=teki2;

    nn=n % 16;
    ang=M_PI/24.0*(double)nn;
    sz=40;
    fl0x=sz*cos(ang-M_PI*2/3);
    fl0y=sz*sin(ang-M_PI*2/3);
    fl1x=sz*cos(ang);
    fl1y=sz*sin(ang);
    fl2x=sz*cos(ang+M_PI*2/3);
    fl2y=sz*sin(ang+M_PI*2/3);
    c0=255-abs(8-nn)*12;
    c1=160-nn*4;
    c2=96+nn*4;

    set_wall(wpt++, fl0x, fl0y,-sz*0.2, 0);
    set_wall(wpt++, fl1x, fl1y,-sz*0.2, 0);
    set_wall(wpt++,    0,    0, sz, 0);
    set_wall(wpt++, 0, 0, c0, 1);
    set_wall(wpt++, fl1x, fl1y,-sz*0.2, 0);
    set_wall(wpt++, fl2x, fl2y,-sz*0.2, 0);
    set_wall(wpt++,    0,    0, sz, 0);
    set_wall(wpt++, 0, 0, c1, 1);
    set_wall(wpt++, fl2x, fl2y,-sz*0.2, 0);
    set_wall(wpt++, fl0x, fl0y,-sz*0.2, 0);
    set_wall(wpt++,    0,    0, sz, 0);
    set_wall(wpt++, 0, 0, c2, 1);

    set_wall(wpt++, fl0x, fl0y,-sz*0.2, 0);
    set_wall(wpt++, fl1x, fl1y,-sz*0.2, 0);
    set_wall(wpt++,    0,    0,-sz*0.5, 0);
    set_wall(wpt++, 0, 0, c0*2/3, 1);
    set_wall(wpt++, fl1x, fl1y,-sz*0.2, 0);
    set_wall(wpt++, fl2x, fl2y,-sz*0.2, 0);
    set_wall(wpt++,    0,    0,-sz*0.5, 0);
    set_wall(wpt++, 0, 0, c1*2/3, 1);
    set_wall(wpt++, fl2x, fl2y,-sz*0.2, 0);
    set_wall(wpt++, fl0x, fl0y,-sz*0.2, 0);
    set_wall(wpt++,    0,    0,-sz*0.5, 0);
    set_wall(wpt++, 0, 0, c2*2/3, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void make_teki3(int n)
{
    double ang;
    double sz;
    int cc;
    int nn;
    struct f_wall *wpt;

    if (n<0){
      teki3=(f_wall *)malloc(sizeof(struct f_wall)*29);
      return;
    }
    wpt=teki3;

    sz=5;
    cc=192;

    set_wall(wpt++,-sz, sz*5,-sz, 0);
    set_wall(wpt++, sz, sz*5,-sz, 0);
    set_wall(wpt++,  0,-sz*5,  0, 0);
    set_wall(wpt++, cc/2, cc/2, 0, 1);
    set_wall(wpt++, sz, sz*5,-sz, 0);
    set_wall(wpt++, sz, sz*5, sz, 0);
    set_wall(wpt++,  0,-sz*5,  0, 0);
    set_wall(wpt++, cc*2/3, cc*2/3, 0, 1);
    set_wall(wpt++, sz, sz*5, sz, 0);
    set_wall(wpt++,-sz, sz*5, sz, 0);
    set_wall(wpt++,  0,-sz*5,  0, 0);
    set_wall(wpt++, cc, cc, 0, 1);
    set_wall(wpt++,-sz, sz*5, sz, 0);
    set_wall(wpt++,-sz, sz*5,-sz, 0);
    set_wall(wpt++,  0,-sz*5,  0, 0);
    set_wall(wpt++, cc*2/3, cc*2/3, 0, 1);

    set_wall(wpt++, sz*0.3,-sz*2, 0, 0);
    set_wall(wpt++, sz*0.7, sz*2, 0, 0);
    set_wall(wpt++, sz*5  , sz*3, 0, 0);
    set_wall(wpt++, cc, cc, 0, 1);
    set_wall(wpt++,-sz*0.3,-sz*2, 0, 0);
    set_wall(wpt++,-sz*0.7, sz*2, 0, 0);
    set_wall(wpt++,-sz*5  , sz*3, 0, 0);
    set_wall(wpt++, cc, cc, 0, 1);
    
    set_wall(wpt++, 0, sz*2, sz*0.7, 0);
    set_wall(wpt++, 0, sz*5, sz    , 0);
    set_wall(wpt++, 0, sz*5, sz*4  , 0);
    set_wall(wpt++, cc/2, cc/2, 0, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void make_teki4(int n)
{
    double ang;
    double fl0x, fl0y, fl1x, fl1y, fl2x, fl2y;
    double fl3x, fl3y, fl4x, fl4y, fl5x, fl5y;
    int c0, c1, c2, c3, c4, c5;
    double sz;
    int nn;
    struct f_wall *wpt;

    if (n<0){
      teki4=(f_wall *)malloc(sizeof(struct f_wall)*41);
      return;
    }
    wpt=teki4;

    nn=n % 16;
    ang=M_PI/24.0*(double)nn;
    sz=40;
    fl0x=sz*0.8*cos(ang-M_PI*2/3);
    fl0y=sz*0.8*sin(ang-M_PI*2/3);
    fl1x=sz*0.8*cos(ang);
    fl1y=sz*0.8*sin(ang);
    fl2x=sz*0.8*cos(ang+M_PI*2/3);
    fl2y=sz*0.8*sin(ang+M_PI*2/3);

    fl3x=sz*0.8*cos(ang-M_PI*1/3);
    fl3y=sz*0.8*sin(ang-M_PI*1/3);
    fl4x=sz*0.8*cos(ang+M_PI*1/3);
    fl4y=sz*0.8*sin(ang+M_PI*1/3);
    fl5x=sz*0.8*cos(ang+M_PI);
    fl5y=sz*0.8*sin(ang+M_PI);

    c0=224-abs(8-nn)*8;
    c1=160-nn*4;
    c2=96+nn*4;
    c3=160+nn*4;
    c4=224-nn*4;
    c5=96+abs(8-nn)*8;

    set_wall(wpt++,fl0x,fl0y, sz*0.5, 0);
    set_wall(wpt++,fl1x,fl1y, sz*0.5, 0);
    set_wall(wpt++,fl2x,fl2y, sz*0.5, 0);
    set_wall(wpt++,255, 0,255, 1);

    set_wall(wpt++,fl0x,fl0y, sz*0.5, 0);
    set_wall(wpt++,fl1x,fl1y, sz*0.5, 0);
    set_wall(wpt++,fl3x,fl3y, sz*0.2, 0);
    set_wall(wpt++, c0, 0, c0, 1);
    set_wall(wpt++,fl1x,fl1y, sz*0.5, 0);
    set_wall(wpt++,fl2x,fl2y, sz*0.5, 0);
    set_wall(wpt++,fl4x,fl4y, sz*0.2, 0);
    set_wall(wpt++, c1, 0, c1, 1);
    set_wall(wpt++,fl2x,fl2y, sz*0.5, 0);
    set_wall(wpt++,fl0x,fl0y, sz*0.5, 0);
    set_wall(wpt++,fl5x,fl5y, sz*0.2, 0);
    set_wall(wpt++, c2, 0, c2, 1);

    set_wall(wpt++,fl0x,fl0y, sz*0.5, 0);
    set_wall(wpt++,fl5x,fl5y, sz*0.2, 0);
    set_wall(wpt++,fl3x,fl3y, sz*0.2, 0);
    set_wall(wpt++, 0, 0, c3, 1);
    set_wall(wpt++,fl1x,fl1y, sz*0.5, 0);
    set_wall(wpt++,fl3x,fl3y, sz*0.2, 0);
    set_wall(wpt++,fl4x,fl4y, sz*0.2, 0);
    set_wall(wpt++, 0, 0, c4, 1);
    set_wall(wpt++,fl2x,fl2y, sz*0.5, 0);
    set_wall(wpt++,fl4x,fl4y, sz*0.2, 0);
    set_wall(wpt++,fl5x,fl5y, sz*0.2, 0);
    set_wall(wpt++, 0, 0, c5, 1);

    set_wall(wpt++,fl3x,fl3y, sz*0.2, 0);
    set_wall(wpt++,fl4x,fl4y, sz*0.2, 0);
    set_wall(wpt++,   0,   0,-sz*0.5, 0);
    set_wall(wpt++, c3*2/3, 0, c3*2/3, 1);
    set_wall(wpt++,fl4x,fl4y, sz*0.2, 0);
    set_wall(wpt++,fl5x,fl5y, sz*0.2, 0);
    set_wall(wpt++,   0,   0,-sz*0.5, 0);
    set_wall(wpt++, c4*2/3, 0, c4*2/3, 1);
    set_wall(wpt++,fl5x,fl5y, sz*0.2, 0);
    set_wall(wpt++,fl3x,fl3y, sz*0.2, 0);
    set_wall(wpt++,   0,   0,-sz*0.5, 0);
    set_wall(wpt++, c5*2/3, 0, c5*2/3, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void make_teki5(int n)
{
    double ang;
    double fl0, fl1;
    int c0, c1, c2, c3;
    double sz;
    int nn;
    struct f_wall *wpt;

    if (n<0){
      teki5=(f_wall *)malloc(sizeof(struct f_wall)*33);
      return;
    }
    wpt=teki5;

    nn=n % 8;
    ang=M_PI/16.0*(double)nn;
    sz=15;
    fl0=sz*cos(ang);
    fl1=sz*sin(ang);
    c0=255-nn*8;
    c1=128-nn*4;
    c2=64+nn*4;
    c3=128+nn*8;

    set_wall(wpt++, fl1,-fl0, sz*0.2, 0);
    set_wall(wpt++, fl0, fl1, sz*0.2, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++, c0/2, c0, c0/2, 1);
    set_wall(wpt++, fl0, fl1, sz*0.2, 0);
    set_wall(wpt++,-fl1, fl0, sz*0.2, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++, c1/2, c1, c1/2, 1);
    set_wall(wpt++,-fl1, fl0, sz*0.2, 0);
    set_wall(wpt++,-fl0,-fl1, sz*0.2, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++, c2/2, c2, c2/2, 1);
    set_wall(wpt++,-fl0,-fl1, sz*0.2, 0);
    set_wall(wpt++, fl1,-fl0, sz*0.2, 0);
    set_wall(wpt++,   0,   0, sz*0.5, 0);
    set_wall(wpt++, c3/2, c3, c3/2, 1);
    set_wall(wpt++, fl1,-fl0, sz*0.2, 0);
    set_wall(wpt++, fl0, fl1, sz*0.2, 0);
    set_wall(wpt++,   0,   0,-sz*0.5, 0);
    set_wall(wpt++, c0/4, c0/2, c0/4, 1);
    set_wall(wpt++, fl0, fl1, sz*0.2, 0);
    set_wall(wpt++,-fl1, fl0, sz*0.2, 0);
    set_wall(wpt++,   0,   0,-sz*0.5, 0);
    set_wall(wpt++, c1/4, c1/2, c1/4, 1);
    set_wall(wpt++,-fl1, fl0, sz*0.2, 0);
    set_wall(wpt++,-fl0,-fl1, sz*0.2, 0);
    set_wall(wpt++,   0,   0,-sz*0.5, 0);
    set_wall(wpt++, c2/4, c2/2, c2/4, 1);
    set_wall(wpt++,-fl0,-fl1, sz*0.2, 0);
    set_wall(wpt++, fl1,-fl0, sz*0.2, 0);
    set_wall(wpt++,   0,   0,-sz*0.5, 0);
    set_wall(wpt++, c3/4, c3/2, c3/4, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void make_building1()
{
    int x, y, n;
    double sz;
    struct f_wall *wpt;

    building1=(f_wall *)malloc(sizeof(struct f_wall)*106);
    wpt=building1;

    sz=8;

    set_wall(&wpt[0],    0,    0, sz*7, 0);
    set_wall(&wpt[1],    0,    0,    0, 0);
    set_wall(&wpt[2], sz*7,    0,    0, 0);
    set_wall(&wpt[3], sz*7,    0, sz*7, 0);
    set_wall(&wpt[4],128,128,128, 1);
    set_wall(&wpt[5], sz*7,    0, sz*7, 0);
    set_wall(&wpt[6], sz*7,    0,    0, 0);
    set_wall(&wpt[7], sz*7, sz*4,    0, 0);
    set_wall(&wpt[8], sz*7, sz*4, sz*7, 0);
    set_wall(&wpt[9], 64, 64, 64, 1);
    set_wall(&wpt[10],    0, sz*4, sz*7, 0);
    set_wall(&wpt[11],    0, sz*4,    0, 0);
    set_wall(&wpt[12], sz*7, sz*4,    0, 0);
    set_wall(&wpt[13], sz*7, sz*4, sz*7, 0);
    set_wall(&wpt[14], 32, 32, 32, 1);
    set_wall(&wpt[15],    0,    0, sz*7, 0);
    set_wall(&wpt[16],    0,    0,    0, 0);
    set_wall(&wpt[17],    0, sz*4,    0, 0);
    set_wall(&wpt[18],    0, sz*4, sz*7, 0);
    set_wall(&wpt[19], 64, 64, 64, 1);
    set_wall(&wpt[20],    0,    0, sz*7, 0);
    set_wall(&wpt[21], sz*7,    0, sz*7, 0);
    set_wall(&wpt[22], sz*7, sz*4, sz*7, 0);
    set_wall(&wpt[23],    0, sz*4, sz*7, 0);
    set_wall(&wpt[24],192,192,192, 1);

    for(y=0;y<2;y++){
      for(x=0;x<2;x++){
        n=25+(y*2+x)*5;
        set_wall(&wpt[n]  ,(x*3+1)*sz,-3,(y*3+3)*sz, 0);
        set_wall(&wpt[n+1],(x*3+1)*sz,-3,(y*3+1)*sz, 0);
        set_wall(&wpt[n+2],(x*3+3)*sz,-3,(y*3+1)*sz, 0);
        set_wall(&wpt[n+3],(x*3+3)*sz,-3,(y*3+3)*sz, 0);
        set_wall(&wpt[n+4], 0,128,128, 1);
      }
    }
    
    set_wall(&wpt[45], 0, 0, 0,-1);
}

void make_building2()
{
    int x, y, n;
    double sz;
    struct f_wall *wpt;

    building2=(f_wall *)malloc(sizeof(struct f_wall)*17);
    wpt=building2;

    sz=30;

    set_wall(wpt++,-sz,-sz,    0, 0);
    set_wall(wpt++, sz,-sz,    0, 0);
    set_wall(wpt++,  0,  0, sz*4, 0);
    set_wall(wpt++,128,128,128, 1);
    set_wall(wpt++, sz,-sz,    0, 0);
    set_wall(wpt++, sz, sz,    0, 0);
    set_wall(wpt++,  0,  0, sz*4, 0);
    set_wall(wpt++,192,192,192, 1);
    set_wall(wpt++, sz, sz,    0, 0);
    set_wall(wpt++,-sz, sz,    0, 0);
    set_wall(wpt++,  0,  0, sz*4, 0);
    set_wall(wpt++,128,128,128, 1);
    set_wall(wpt++,-sz, sz,    0, 0);
    set_wall(wpt++,-sz,-sz,    0, 0);
    set_wall(wpt++,  0,  0, sz*4, 0);
    set_wall(wpt++, 64, 64, 64, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void make_building3()
{
    int x, y, n;
    double sz;
    struct f_wall *wpt;

    building3=(f_wall *)malloc(sizeof(struct f_wall)*23);
    wpt=building3;

    sz=20;

    set_wall(wpt++,-sz,-sz, 256, 0);
    set_wall(wpt++,-sz,-sz,   0, 0);
    set_wall(wpt++, sz,-sz,   0, 0);
    set_wall(wpt++, sz,-sz, 256, 0);
    set_wall(wpt++,128,128,128, 1);
    set_wall(wpt++, sz,-sz, 256, 0);
    set_wall(wpt++, sz,-sz,   0, 0);
    set_wall(wpt++, sz, sz,   0, 0);
    set_wall(wpt++, sz, sz, 256, 0);
    set_wall(wpt++, 64, 64, 64, 1);
    set_wall(wpt++,-sz, sz, 256, 0);
    set_wall(wpt++,-sz, sz,   0, 0);
    set_wall(wpt++, sz, sz,   0, 0);
    set_wall(wpt++, sz, sz, 256, 0);
    set_wall(wpt++, 32, 32, 32, 1);
    set_wall(wpt++,-sz,-sz, 256, 0);
    set_wall(wpt++,-sz,-sz,   0, 0);
    set_wall(wpt++,-sz, sz,   0, 0);
    set_wall(wpt++,-sz, sz, 256, 0);
    set_wall(wpt++, 64, 64, 64, 1);
    set_wall(wpt++, 0, 0, 0,-1);
}

void make_shot()
{
    double fl0x, fl0y, fl1x, fl1y, fl2x, fl2y;
    struct f_wall *wpt;

    shot=(f_wall *)malloc(sizeof(struct f_wall)*17);
    wpt=shot;

    fl0x=10.0*cos(-M_PI*2/3);
    fl0y=10.0*sin(-M_PI*2/3);
    fl1x=10.0*cos(0);
    fl1y=10.0*sin(0);
    fl2x=10.0*cos(M_PI*2/3);
    fl2y=10.0*sin(M_PI*2/3);

    set_wall(wpt++, fl0x,  0, fl0y, 0);
    set_wall(wpt++, fl1x,  0, fl1y, 0);
    set_wall(wpt++, 0.01, 30,-0.01, 0);
    set_wall(wpt++, 64, 64, 64, 1);
    set_wall(wpt++, fl1x,  0, fl1y, 0);
    set_wall(wpt++, fl2x,  0, fl2y, 0);
    set_wall(wpt++, 0.01, 30, 0.01, 0);
    set_wall(wpt++,128,128,128, 1);
    set_wall(wpt++, fl2x,  0, fl2y, 0);
    set_wall(wpt++, fl0x,  0, fl0y, 0);
    set_wall(wpt++,-0.01, 30,    0, 0);
    set_wall(wpt++, 32, 32, 32, 1);
    set_wall(wpt++, fl0x,  0, fl0y, 0);
    set_wall(wpt++, fl1x,  0, fl1y, 0);
    set_wall(wpt++, fl2x,  0, fl2y, 0);
    set_wall(wpt++,255, 64, 64, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void make_bomb()
{
    double fl0x, fl0y, fl1x, fl1y, fl2x, fl2y;
    struct f_wall *wpt;

    bomb=(f_wall *)malloc(sizeof(struct f_wall)*13);
    wpt=bomb;

    fl0x=10.0*cos(-M_PI*2/3);
    fl0y=10.0*sin(-M_PI*2/3);
    fl1x=10.0*cos(0);
    fl1y=10.0*sin(0);
    fl2x=10.0*cos(M_PI*2/3);
    fl2y=10.0*sin(M_PI*2/3);

    set_wall(wpt++, fl0x,  0, fl0y, 0);
    set_wall(wpt++, fl1x,  0, fl1y, 0);
    set_wall(wpt++, 0.01,-30,-0.01, 0);
    set_wall(wpt++,192,192,192, 1);
    set_wall(wpt++, fl1x,  0, fl1y, 0);
    set_wall(wpt++, fl2x,  0, fl2y, 0);
    set_wall(wpt++, 0.01,-30, 0.01, 0);
    set_wall(wpt++,255,255,255, 1);
    set_wall(wpt++, fl2x,  0, fl2y, 0);
    set_wall(wpt++, fl0x,  0, fl0y, 0);
    set_wall(wpt++,-0.01,-30,    0, 0);
    set_wall(wpt++,128,128,128, 1);
    set_wall(wpt++,  0,  0,  0,-1);
}

void show_title(int num)
{
  struct f_vec my_pos;
  struct f_vec my_eye;

  tft.clearDisplay();

  my_pos.x=128;
  my_pos.y=0;
  my_pos.z=128;
  my_eye.x=0;
  my_eye.y=1;
  my_eye.z=0;

  push_polygon(NULL,0,0,0,0);

  draw_wall(0,  0,  8,  0,0,&my_pos,&my_eye);
  draw_wall(9,128,300,  0,0,&my_pos,&my_eye);
  draw_wall(8,  0,200,  0,0,&my_pos,&my_eye);
  draw_wall(8,186,200,  0,0,&my_pos,&my_eye);
  draw_wall(1, 64,150,192,0,&my_pos,&my_eye);
  draw_wall(2,192,150,192,0,&my_pos,&my_eye);
  draw_wall(3, 64,120, 64,0,&my_pos,&my_eye);
  draw_wall(4,192,150, 64,0,&my_pos,&my_eye);

  draw_polygon();

  tft.fillRect(0, 0, _width, SCREEN_START_Y, ILI9163_BLACK);
  tft.fillRect(0, 0, SCREEN_START_X, _height, ILI9163_BLACK);
  tft.drawRect(SCREEN_START_X, SCREEN_START_Y, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, ILI9163_WHITE);

  tft.setTextSize(2);
  tft.setTextColor(ILI9163_WHITE);
  tft.setCursor(8, 8);
  tft.print("3D shoot");

  tft.display();
}

void show_game(int num)
{
  int i;
  struct f_vec my_pos;
  struct f_vec my_eye;

  tft.clearDisplay();

  my_pos.x=my_x;
  my_pos.y=0;
  my_pos.z=my_y;
  my_eye.x=0;
  my_eye.y=1;
  my_eye.z=0;

  push_polygon(NULL,0,0,0,0);

  draw_wall(0, 0, 8, 0,0,&my_pos,&my_eye);

  // Show teki
  for(i=0;i<MAX_TEKI;i++){
    if (teki_x[i]>=0)
      draw_wall(teki_typ[i],
                teki_x[i],teki_y[i],teki_z[i],0,&my_pos,&my_eye);
  }

  // Show bomb
  for(i=0;i<MAX_BOMB;i++){
    if (bomb_x[i]>=0)
      draw_wall(14,bomb_x[i],my_pos.y+bomb_y[i],bomb_z[i],0,&my_pos,&my_eye);
  }

  // Show shot
  for(i=0;i<MAX_SHOT;i++){
    if (shot_x[i]>=0)
      draw_wall(15,shot_x[i],my_pos.y+shot_y[i],shot_z[i],0,&my_pos,&my_eye);
  }

  draw_polygon();

  tft.fillRect(0, 0, _width, SCREEN_START_Y, ILI9163_BLACK);
  tft.fillRect(0, 0, SCREEN_START_X, _height, ILI9163_BLACK);
  tft.drawRect(SCREEN_START_X, SCREEN_START_Y, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, ILI9163_WHITE);

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE);
  tft.setCursor(_width-64, 8);
  tft.print(score);

  tft.display();
}

void show_clear()
{
  int i;
  char str[32];
  struct f_vec my_pos;
  struct f_vec my_eye;

  tft.clearDisplay();

  my_pos.x=my_x;
  my_pos.y=0;
  my_pos.z=my_y;
  my_eye.x=0;
  my_eye.y=1;
  my_eye.z=0;

  push_polygon(NULL,0,0,0,0);

  draw_wall(0, 0, 8, 0,0,&my_pos,&my_eye);

  draw_polygon();

  tft.fillRect(0, 0, _width, SCREEN_START_Y, ILI9163_BLACK);
  tft.fillRect(0, 0, SCREEN_START_X, _height, ILI9163_BLACK);
  tft.drawRect(SCREEN_START_X, SCREEN_START_Y, SCREEN_WIDTH-1, SCREEN_HEIGHT-1, ILI9163_WHITE);

  tft.setTextSize(1);
  tft.setTextColor(ILI9163_WHITE);
  tft.setCursor(_width-64, 8);
  tft.print(score);

  tft.setTextSize(2);
  tft.setTextColor(ILI9163_CYAN);
  sprintf(str, "STAGE %d", stage);
  tft.setCursor(_width/2-36, _height/2-24);
  tft.print(str);
  
  tft.setCursor(_width/2-36, _height/2+8);
  tft.print("Clear !");

  tft.display();
}

void show_gameover()
{
  tft.setTextSize(2);
  tft.setTextColor(ILI9163_RED);
  tft.setCursor(_width/2-54, _height/2-8);
  tft.print("Game over");

  tft.display();
}

void draw_wall(int no,int x,int y,int z,int dir,struct f_vec *pos,struct f_vec *eye)
{
  int i;
  struct f_wall *tmppt;
  struct f_vec wall_xyz;
  struct f_vec vis_pos;
  struct f_vec tmp;
  int back;
  double dirmax,dirmin;
  POINT aPoints[8];

  tmppt=wallpt[no];
  while(tmppt->typ>=0){
    back=0;
    dirmax=0;
    dirmin=10000000;
    for(i=0;tmppt->typ==0;i++){
    if(dir==0){
      wall_xyz.x=(double)(tmppt->x+x);
      wall_xyz.y=(double)(tmppt->y+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    if(dir==1){
      wall_xyz.x=(double)(-tmppt->y+x);
      wall_xyz.y=(double)(tmppt->x+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    if(dir==2){
      wall_xyz.x=(double)(-tmppt->x+x);
      wall_xyz.y=(double)(-tmppt->y+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    if(dir==3){
      wall_xyz.x=(double)(tmppt->y+x);
      wall_xyz.y=(double)(-tmppt->x+y);
      wall_xyz.z=(double)(tmppt->z+z);
    }
    pos3d(&vis_pos,eye,vec_sub(&tmp,&wall_xyz,pos));
    aPoints[i].x = (int)(vis_pos.x*SCREEN_WIDTH/2)+SCREEN_WIDTH/2;
    aPoints[i].y = (int)(-vis_pos.y*SCREEN_HEIGHT/2)+SCREEN_HEIGHT/2;
    if(vis_pos.z<0)
      back=1;
    if(vis_pos.z>dirmax)
      dirmax=vis_pos.z;
    if(vis_pos.z<dirmin)
      dirmin=vis_pos.z;
    tmppt++;
    }
    if(back==0){
      if(tmppt->typ==1){
        push_polygon(aPoints,i,tft.color565(tmppt->x,tmppt->y,tmppt->z),dirmax,dirmin);
      }
    }
    tmppt++;
  }
  return;
}

void push_polygon(POINT *pntpt, int num, uint16_t col, double dirmax, double dirmin)
{
  int i;
  if(num==0){
    poly_num=0;
    return;
  }
  for(i=0;i<num;i++){
    polygon_dat[poly_num].pnt[i].x=pntpt[i].x;
    polygon_dat[poly_num].pnt[i].y=pntpt[i].y;
  }
  polygon_dat[poly_num].n=num;
  polygon_dat[poly_num].c=col;
  polygon_dat[poly_num].dmax=dirmax;
  polygon_dat[poly_num].dmin=dirmin;
  poly_num++;
  return;
}

void draw_polygon()
{
  int i,j;
  int num;
  double dirmax,dirmin;

  for(i=0;i<poly_num;i++){
    dirmax=0;
    dirmin=0;
    for(j=0;j<poly_num;j++){
      if(polygon_dat[j].n>0){
/*        
        if(polygon_dat[j].dmin>dirmin){
          num=j;
          dirmax=polygon_dat[j].dmax;
          dirmin=polygon_dat[j].dmin;
        }else if(polygon_dat[j].dmin==dirmin && polygon_dat[j].dmax>dirmax){
          num=j;
          dirmax=polygon_dat[j].dmax;
          dirmin=polygon_dat[j].dmin;
        }
*/
        if(polygon_dat[j].dmax>dirmax){
          num=j;
          dirmax=polygon_dat[j].dmax;
          dirmin=polygon_dat[j].dmin;
        }else if(polygon_dat[j].dmax==dirmax && polygon_dat[j].dmin>dirmin){
          num=j;
          dirmax=polygon_dat[j].dmax;
          dirmin=polygon_dat[j].dmin;
        }
      }
    }
    if(dirmax==0)
      return;

    Polygon(polygon_dat[num].pnt, polygon_dat[num].n, polygon_dat[num].c);
    polygon_dat[num].n=0;
  }
  return;
}

void Polygon(POINT *pnt, int n, uint16_t col)
{
  int i;

  if (n<3)
    return;
  for(i=2; i<n; i++)
    tft.fillTriangle(SCREEN_START_X+pnt[0].x, SCREEN_START_Y+pnt[0].y,
                     SCREEN_START_X+pnt[i-1].x, SCREEN_START_Y+pnt[i-1].y,
                     SCREEN_START_X+pnt[i].x, SCREEN_START_Y+pnt[i].y, col);
}

struct f_vec *pos3d(struct f_vec *vis,struct f_vec *eye,struct f_vec *obj)
{
  double angx,angy;
  double lenga,lengb,lengc;
  int back;
  back=0;
  lenga=sqrt(eye->x*eye->x+eye->y*eye->y);
  lengb=sqrt(obj->x*obj->x+obj->y*obj->y);
  if(lengb==0)
    angx=0;
  else{
    angx=(eye->x*obj->x+eye->y*obj->y)/lenga/lengb;
    if(angx>=1.0)
    angx=1.0;
    if(angx<=-1.0)
    angx=-1.0;
    angx=acos(angx);
    if(angx>=(M_PI/2-0.01)){
    back=1;
    vis->x=0;
    vis->y=0;
    vis->z=-1;
    return(vis);
    }
  }
  if((-eye->x*obj->y+eye->y*obj->x)<0)
    angx=-angx;
  lengc=sqrt(eye->x*eye->x+eye->y*eye->y);
  lengc=(obj->x*eye->x+obj->y*eye->y)/lengc;
  if(lengc==0){
    if(obj->z>0)
    angy=1;
    else
    angy=-1;
  }else
    angy=obj->z/fabs(lengc);
  vis->x=tan(angx);
  vis->y=angy;
  vis->z=obj->x*obj->x+obj->y*obj->y+obj->z*obj->z;
  return(vis);
}

struct f_vec *vec_add(struct f_vec *result,struct f_vec *para0,struct f_vec *para1)
{
  result->x=para0->x+para1->x;
  result->y=para0->y+para1->y;
  result->z=para0->z+para1->z;
  return(result);
}

struct f_vec *vec_sub(struct f_vec *result,struct f_vec *para0,struct f_vec *para1)
{
  result->x=para0->x-para1->x;
  result->y=para0->y-para1->y;
  result->z=para0->z-para1->z;
  return(result);
}
/*
void tone (uint8_t pin, uint16_t freq, uint16_t leng)
{
  if (tone_init==0){
    ledcSetup(1, 12000, 8);
    ledcAttachPin(pin, 1);
    tone_init=1;
  }
  ledcWriteTone(1, freq);
}

void noTone (uint8_t pin)
{
  ledcWriteTone(1, 1);
}
*/
void i2s_init()
{
    i2s_config_t i2s_config = {
    .mode =  (i2s_mode_t)(I2S_MODE_MASTER | I2S_MODE_TX  | I2S_MODE_RX),
    .sample_rate = fs,
    .bits_per_sample = (i2s_bits_per_sample_t)32,
    .channel_format = I2S_CHANNEL_FMT_RIGHT_LEFT,
    .communication_format = (i2s_comm_format_t)(I2S_COMM_FORMAT_STAND_I2S),
    .intr_alloc_flags = 0,
    .dma_buf_count = 6,
    .dma_buf_len = BLOCK_SAMPLES*4,
    .use_apll = false,
    .tx_desc_auto_clear = true,
    .fixed_mclk = 0,
  };
  i2s_driver_install( I2S_NUM_0, &i2s_config, 0, NULL);

  i2s_pin_config_t pin_config = {
        .bck_io_num   = 18,
        .ws_io_num    = 19,
        .data_out_num = 8,
        .data_in_num  = 1                                                       
    };
  i2s_set_pin( I2S_NUM_0, &pin_config);
  i2s_zero_dma_buffer(I2S_NUM_0);
}
void i2s_tone (uint8_t pin, uint16_t freq, uint16_t leng)
{
  int i;
  int j;
  int ll;
  float ff;
  size_t writesize = 0;

  ll=leng*fs/BLOCK_SAMPLES/1000;
  if (ll<1)
    ll=1;
  if (ll>8)
    ll=8;
  ff=2*3.14159*(float)freq/fs;
  for (i=0; i<BLOCK_SAMPLES*ll; i++) {
    Lch_out[i] = (float)16384.*sin(ff*(float)i);
    Rch_out[i] = (float)16384.*sin(ff*(float)i);
  }

  //Output to I2S codec
  for(j=0;j<ll;j++){
    for (i=0; i<BLOCK_SAMPLES; i++) {
      txbuf[i*2] = (int) Lch_out[BLOCK_SAMPLES*j+i];
      txbuf[i*2+1] = (int) Rch_out[BLOCK_SAMPLES*j+i];
    }
    i2s_write( I2S_NUM_0, &txbuf[0], BLOCK_SAMPLES*2*4, &writesize, portMAX_DELAY);              
  }

  for (i=0; i<BLOCK_SAMPLES; i++) {
    txbuf[i*2] = 0;
    txbuf[i*2+1] = 0;
  }
  i2s_write( I2S_NUM_0, &txbuf[0], BLOCK_SAMPLES*2*4, &writesize, portMAX_DELAY);              
}

void i2s_noTone (uint8_t pin)
{
  int i;
  size_t writesize = 0;

  for (i=0; i<BLOCK_SAMPLES; i++) {
    txbuf[i*2] = 0;
    txbuf[i*2+1] = 0;
  }
  i2s_write( I2S_NUM_0, &txbuf[0], BLOCK_SAMPLES*2*4, &writesize, portMAX_DELAY);              
}
